//+=============================================================================
//
// file :         SubscribeThread.cpp
//
// description :  C++ source for thread management
// project :      TANGO Device Server
//
// $Author: graziano $
//
//
//
//
// copyleft :     European Synchrotron Radiation Facility
//                BP 220, Grenoble 38043
//                FRANCE
//
//-=============================================================================


#include "SubscribeThread.h"
#include "PushThread.h"
#include <AlarmNotify.h>
#include "AlarmSignal.h"

namespace AlarmNotify_ns
{
    SharedData::SharedData(Tango::DeviceImpl *dev): Tango::LogAdapter(dev)
                                            , initialized(false)
                                            , init_condition(&init_mutex)
    {
        _device=dev;
        stop_it=false;
    }

    SharedData::~SharedData()
    {
        init_condition.broadcast();
    }

    //=============================================================================
    /**
     * get signal by name.
     */
    //=============================================================================
    auto SharedData::get_signal(const string &signame) -> std::shared_ptr<AlarmSignal>
    {
        ReaderLock lock(veclock);
        //omni_mutex_lock sync(*this);
        for (auto &signal : signals)
        {
            if(is_same_signal_name(signal->name, signame))
                return signal;
        }

        Tango::Except::throw_exception(
                (const char *)"BadSignalName",
                "Signal " + signame + " NOT FOUND in signal list",
                (const char *)"SharedData::get_signal()");
    }

    //=============================================================================
    /**
     * Check if two strings represents the same signal.
     */
    //=============================================================================
    auto SharedData::is_same_signal_name(const std::string& name1, const std::string& name2) -> bool
    {
        // Easy case
        if(name1 == name2)
            return true;

#ifndef _MULTI_TANGO_HOST
        if(AlarmStringUtils::compare_without_domain(name1, name2))
#else  
            if(!hdb_dev->compare_tango_names(name1, name2))
#endif
                return true;

        return false;
    }

    //=============================================================================
    /**
     * Remove a signal in the list.
     */
    //=============================================================================
    void SharedData::remove(const string& signame, bool stop)
    {
        // Remove in signals list (vector)
        auto sig = get_signal(signame);
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        
        if(stop)
        {
            unsubscribe_events(sig);
        }
        else
        {
            size_t idx = 0;
            bool running = false;
            bool stopped = false;
            {
                WriterLock lock(veclock);
                auto pos = signals.begin();

                for(size_t i = 0; i < signals.size(); ++i, pos++)
                {
                    std::shared_ptr<AlarmSignal> sig = signals[i];
                    if(is_same_signal_name(sig->name, signame))
                    {
                        // devalidate its index in case it tries to access one of the list
                        running = sig->is_running();
                        stopped = sig->is_stopped();
                        
                        signals.erase(pos);
                        DEBUG_STREAM <<"SharedData::"<< __func__<<": removed " << signame << endl;
                        break;
                    }
                }
            }
            
            _dev->_remove_attribute(idx, running, stopped);

            signal();
        }
    }
    //=============================================================================
    /**
     * Unsubscribe events for a signal
     */
    //=============================================================================
    void SharedData::unsubscribe_events(std::shared_ptr<AlarmSignal> sig)
    {
        std::string signame = sig->name;
        try
        {
            DEBUG_STREAM <<"SharedData::"<< __func__<<": unsubscribing events... "<< signame << endl;
            sig->remove_callback();
            DEBUG_STREAM <<"SharedData::"<< __func__<<": unsubscribed events... "<< signame << endl;
        }
        catch (Tango::DevFailed &e)
        {
            // Do nothing
            // Unregister failed means Register has also failed
            //sig->siglock->writerIn();
            INFO_STREAM <<"SharedData::"<< __func__<<": Exception unsubscribing " << signame << " err=" << e.errors[0].desc << endl;
        }
    }
    //=============================================================================
    /**
     * Start saving on DB a signal.
     */
    //=============================================================================
    void SharedData::start(const string& signame)
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        auto signal = get_signal(signame);

        //_dev->push_thread->start_attr(*signal);

        if(!signal->is_running())
        {
            if(signal->is_stopped())
            {
                _dev->AttributeStoppedNumber--;
                add(signal, true);
            }

            _dev->AttributeStartedNumber++;

            signal->set_running();
        }
    }
    //=============================================================================
    /**
     * Stop saving on DB a signal.
     */
    //=============================================================================
    void SharedData::stop(const string& signame)
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        auto signal = get_signal(signame);

        //_dev->push_thread->stop_attr(*signal);

        if(!signal->is_stopped())
        {
            _dev->AttributeStoppedNumber++;
            if(signal->is_running())
            {
                _dev->AttributeStartedNumber--;
                try
                {
                    unsubscribe_events(signal);
                }
                catch (Tango::DevFailed &e)
                {
                    //Tango::Except::print_exception(e);
                    INFO_STREAM << "SharedData::stop: error removing  " << signame << endl;
                }
            }

            signal->set_stopped();
        }
    }
    //=============================================================================
    /**
     * Start saving on DB all signals.
     */
    //=============================================================================
    void SharedData::start_all()
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        ReaderLock lock(veclock);
        for (auto &signal : signals)
        {
            signal->set_running();
        }
        _dev->AttributeStoppedNumber=0;
        _dev->AttributeStartedNumber=signals.size();
    }
    //=============================================================================
    /**
     * Stop saving on DB all signals.
     */
    //=============================================================================
    void SharedData::stop_all()
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        ReaderLock lock(veclock);
        for (auto &signal : signals)
        {
            signal->set_stopped();
        }
        _dev->AttributeStoppedNumber=signals.size();
        _dev->AttributeStartedNumber=0;
    }
    //=============================================================================
    /**
     * Is a signal saved on DB?
     */
    //=============================================================================
    auto SharedData::is_running(const string& signame) -> bool
    {
        //to be locked if called outside lock in ArchiveCB::push_event

        auto signal = get_signal(signame);

        return signal->is_running();
    }
    //=============================================================================
    /**
     * Is a signal not subscribed?
     */
    //=============================================================================
    auto SharedData::is_stopped(const string& signame) -> bool
    {
        //to be locked if called outside lock in ArchiveCB::push_event

        auto signal = get_signal(signame);

        return signal->is_stopped();
    }
    //=============================================================================
    /**
     * Is a signal first event arrived?
     */
    //=============================================================================
    auto SharedData::is_first(const string& signame) -> bool
    {
        auto signal = get_signal(signame);

        return signal->is_first();
    }
    //=============================================================================
    /**
     * Set a signal first event arrived
     */
    //=============================================================================
    void SharedData::set_first(const string& signame)
    {
        auto signal = get_signal(signame);

        signal->set_first();
    }
    //=============================================================================
    /**
     * Is a signal first consecutive error event arrived?
     */
    //=============================================================================
    auto SharedData::is_first_err(const string& signame) -> bool
    {
        auto signal = get_signal(signame);

        return signal->is_first_err();
    }
    //=============================================================================
    /**
     * Remove a signal in the list.
     */
    //=============================================================================
    void SharedData::clear_signals()
    {
        DEBUG_STREAM <<"SharedData::"<<__func__<< "    entering..."<< endl;
        veclock.readerIn();
        vector<std::shared_ptr<AlarmSignal>> local_signals(signals);
        veclock.readerOut();

        {
            WriterLock lock(veclock);
            signals.clear();
        }

        for (const auto &signal : local_signals)
        {
            signal->remove_callback();
        }
        DEBUG_STREAM <<"SharedData::"<< __func__<< ": exiting..."<<endl;
    }
    //=============================================================================
    /**
     * Add a new signal.
     */
    //=============================================================================
    void SharedData::add(const string& signame)
    {
        //AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        auto signal = add(signame, false);
        add(signal, true);
    }
    //=============================================================================
    /**
     * Add a new signal.
     */
    //=============================================================================
    void SharedData::add(std::shared_ptr<AlarmSignal> sig, bool start)
    {
        sig->init();
        if(start)
        {
            sig->start();
        }

        DEBUG_STREAM <<"SubscribeThread::"<< __func__<<": exiting... " << sig->name << endl;
        signal();
    }

    std::shared_ptr<AlarmSignal> SharedData::add(const string& signame, bool start)
    {
        //DEBUG_STREAM << "SharedData::"<<__func__<<": Adding " << signame << " start="<<(start ? "Y" : "N")<< endl;

        // Check if already subscribed
        bool found = false;
        std::shared_ptr<AlarmSignal> signal;
        try
        {
            signal = get_signal(signame);
            found = true;
        }
        catch(Tango::DevFailed &e)
        {
        }
        //DEBUG_STREAM << "SharedData::"<<__func__<<": signame="<<signame<<" found="<<(found ? "Y" : "N") << " start="<<(start ? "Y" : "N")<< endl;
        if (found && !start)
            Tango::Except::throw_exception(
                    (const char *)"BadSignalName",
                    "Signal " + signame + " already subscribed",
                    (const char *)"SharedData::add()");

        if (!found)
        {
            DEBUG_STREAM << "SharedData::"<<__func__<<": Adding " << signame << " start="<<(start ? "Y" : "N")<< endl;
            signal = std::make_shared<AlarmSignal>(_device, signame);
            AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
            veclock.writerIn();
            // Add in vector
            signals.push_back(signal);
            _dev->AttributeNumber++;

            if(!start)
                _dev->AttributeStoppedNumber++;

            veclock.writerOut();
        }

        return signal;

        //condition.signal();
    }
    //=============================================================================
    /**
     * Subscribe archive event for each signal
     */
    //=============================================================================
    void SharedData::subscribe_events()
    {
        /*for (unsigned int ii=0 ; ii<signals.size() ; ii++)
          {
          AlarmSignal *sig2 = &signals[ii];
          int ret = pthread_rwlock_trywrlock(&sig2->siglock);
          DEBUG_STREAM << __func__<<": pthread_rwlock_trywrlock i="<<ii<<" name="<<sig2->name<<" just entered " << ret << endl;
          if(ret == 0) pthread_rwlock_unlock(&sig2->siglock);
          }*/
        //omni_mutex_lock sync(*this);
        veclock.readerIn();
        for (const auto& sig : signals)
        {
            if (sig->is_not_subscribed())
            {
                try
                {
                    add(sig, true);

                }
                catch (Tango::DevFailed &e)
                {
                    //Tango::Except::print_exception(e);
                    INFO_STREAM << "SharedData::subscribe_events: error adding  " << sig->name <<" err="<< e.errors[0].desc << endl;
                }
                sig->subscribe_events(_device);
            }
        }
        veclock.readerOut();

        init_mutex.lock();
        initialized = true;
        init_mutex.unlock();

        init_condition.signal();
    }
    //=============================================================================
    //=============================================================================
    auto SharedData::is_initialized() -> bool
    {
        bool ret;
        init_mutex.lock();
        ret = initialized;
        init_mutex.unlock();
        return ret; 
    }
    //=============================================================================
    /**
     * return number of signals to be subscribed
     */
    //=============================================================================
    auto SharedData::nb_sig_to_subscribe() -> int
    {
        ReaderLock lock(veclock);

        int nb = 0;
        for(const auto& signal : signals)
        {
            if(signal->is_not_subscribed())
            {
                ++nb;
            }
        }
        return nb;
    }
    //=============================================================================
    /**
     * Return the list of signals
     */
    //=============================================================================
    void SharedData::get_sig_list(vector<string> &list)
    {
        ReaderLock lock(veclock);
        list.clear();
        for(const auto& signal : signals)
        {
            string signame(signal->name);
            list.push_back(signame);
        }
    }
    //=============================================================================
    /**
     * Return the list of errors
     */
    //=============================================================================
    auto SharedData::get_error_list(vector<string> &list) -> bool
    {
        bool changed=false;
        ReaderLock lock(veclock);
        size_t old_size = list.size();
        size_t i = 0;
        for (i=0 ; i<signals.size() && i < old_size ; i++)
        {
            string err = signals[i]->get_error();

            if(err != list[i])
            {
                list[i] = err;
                changed = true;
            }
        }
        if(signals.size() < old_size)
        {
            list.erase(list.begin()+i, list.begin()+old_size);
            changed = true;
        }
        else
        {
            for (size_t i=old_size ; i<signals.size() ; i++)
            {
                string err = signals[i]->get_error();

                list.push_back(err);
                changed = true;
            }
        }
        return changed;
    }
    //=============================================================================
    /**
     * Return the list of errors
     */
    //=============================================================================
    void SharedData::get_ev_counter_list(vector<uint32_t> &list)
    {
        ReaderLock lock(veclock);
        list.clear();
        for(const auto& signal : signals)
        {
            list.push_back(signal->get_total_events());
        }
    }
    //=============================================================================
    /**
     * Return the complete, started and stopped lists of signals
     */
    //=============================================================================
    auto SharedData::get_lists(vector<string> &s_list, vector<string> &s_start_list, vector<string> &s_stop_list) -> bool
    {
        bool changed = false;
        ReaderLock lock(veclock);
        size_t old_s_list_size = s_list.size();
        size_t old_s_start_list_size = s_start_list.size();
        size_t old_s_stop_list_size = s_stop_list.size();
        vector<string> tmp_start_list;
        vector<string> tmp_stop_list;
        //update list
        size_t i = 0;
        for (i=0 ; i<signals.size() && i< old_s_list_size; i++)
        {
            string signame(signals[i]->name);
            if(signame != s_list[i])
            {
                s_list[i] = signame;
                changed = true;
            }
        }

        if(signals.size() < old_s_list_size)
        {
            s_list.erase(s_list.begin()+i, s_list.begin()+old_s_list_size);
            changed = true;
        }
        else
        {
            for (i=old_s_list_size ; i<signals.size() ; i++)
            {
                changed = true;
                string signame(signals[i]->name);
                s_list.push_back(signame);
            }
        }

        for(const auto& signal : signals)
        {
            string signame(signal->name);
            if (signal->is_running())
            {
                tmp_start_list.push_back(signame);
            }
            else if(signal->is_stopped())
            {
                tmp_stop_list.push_back(signame);
            }
        }
        //update start list
        for (i=0 ; i<tmp_start_list.size() && i< old_s_start_list_size; i++)
        {
            string signame(tmp_start_list[i]);
            if(signame != s_start_list[i])
            {
                s_start_list[i] = signame;
                changed = true;
            }
        }
        if(tmp_start_list.size() < old_s_start_list_size)
        {
            s_start_list.erase(s_start_list.begin()+i, s_start_list.begin()+old_s_start_list_size);
            changed = true;
        }
        else
        {
            for (size_t i=old_s_start_list_size ; i<tmp_start_list.size() ; i++)
            {
                changed = true;
                string signame(tmp_start_list[i]);
                s_start_list.push_back(signame);
            }
        }
        //update stop list
        for (i=0 ; i<tmp_stop_list.size() && i< old_s_stop_list_size; i++)
        {
            string signame(tmp_stop_list[i]);
            if(signame != s_stop_list[i])
            {
                s_stop_list[i] = signame;
                changed = true;
            }
        }
        if(tmp_stop_list.size() < old_s_stop_list_size)
        {
            s_stop_list.erase(s_stop_list.begin()+i, s_stop_list.begin()+old_s_stop_list_size);
            changed = true;
        }
        else
        {
            for (size_t i=old_s_stop_list_size ; i<tmp_stop_list.size() ; i++)
            {
                changed = true;
                string signame(tmp_stop_list[i]);
                s_stop_list.push_back(signame);
            }
        }
        return changed;
    }

    //=============================================================================
    /**
     * Return the list of signals + status + freq
     */
    //=============================================================================
	void SharedData::get_sig_list_status(vector<string> &s_list)
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        for(const auto& signal : signals)
        {
            ostringstream sig_status;
            string signame(signal->name);
            sig_status << signal->name;
#if 0
            if (signal->is_running())
            {
                sig_status << ", RUNNING";
            }
            else if(signal->is_paused())
            {
                sig_status << ", PAUSED";
            }
            else if(signal->is_stopped())
            {
                sig_status << ", STOPPED";
            }
#endif
            sig_status << ", " << (signal->is_on_error() ? "ERROR" : signal->get_alarm_state());
            sig_status << ", counter=" << signal->get_ok_notification();
            auto rate = signal->get_rate_ok_notification(_dev->alarmFloodPeriod);
            sig_status << ", rate=" << rate;
            if(rate > _dev->alarmFloodThresh)
                sig_status << " FLOOD";
            s_list.push_back(sig_status.str());
        }        
    }

    //=============================================================================
    /**
     * Get the ok counter of event rx
     */
    //=============================================================================
    auto SharedData::get_ok_event(const string& signame) -> uint32_t
    {
        auto signal = get_signal(signame);

        return signal->get_ok_events();
    }
    //=============================================================================
    /**
     * Get last okev timestamp
     */
    //=============================================================================
    auto SharedData::get_last_okev(const string &signame) -> std::chrono::time_point<std::chrono::system_clock>
    {
        auto signal = get_signal(signame);

        return signal->get_last_ok_event();
    }
    //=============================================================================
    /**
     * Get the error counter of event rx
     */
    //=============================================================================
    auto SharedData::get_nok_event(const string& signame) -> uint32_t
    {
        auto signal = get_signal(signame);

        return signal->get_nok_events();
    }
    //=============================================================================
    /**
     * Get last nokev timestamp
     */
    //=============================================================================
    auto SharedData::get_last_nokev(const string& signame) -> std::chrono::time_point<std::chrono::system_clock>
    {
        auto signal = get_signal(signame);

        return signal->get_last_nok_event();    
    }

    //=============================================================================
    /**
     *	Get the error counter of db saving
     */
    //=============================================================================
    auto SharedData::get_nok_notification(const string &signame) -> uint32_t
    {
        auto signal = get_signal(signame);

        return signal->get_nok_notification();
    }
    //=============================================================================
    /**
     *	Get last nokdb timestamp
     */
    //=============================================================================
    auto SharedData::get_last_nokdb(const string &signame) -> std::chrono::time_point<std::chrono::system_clock>
    {
        auto signal = get_signal(signame);

        return signal->get_last_nok_notification();    
    }
    //=============================================================================
    /**
     * Reset statistic counters
     */
    //=============================================================================
    void SharedData::reset_statistics()
    {
        ReaderLock lock(veclock);
        for (auto &signal : signals)
        {
            signal->reset_statistics();
        }
    }
    //=============================================================================
    /**
     * Return ALARM if at list one signal is not subscribed.
     */
    //=============================================================================
    auto SharedData::state() -> Tango::DevState
    {
        ReaderLock lock(veclock);
        Tango::DevState state = Tango::ON;
        for (const auto &sig : signals)
        {
            if (sig->get_state() == Tango::ALARM && sig->is_running())
            {
                state = Tango::ALARM;
                break;
            }
        }
        return state;
    }
    //=============================================================================
    //=============================================================================
    auto SharedData::get_if_stop() -> bool
    {
        omni_mutex_lock sync(*this);
        return stop_it;
    }
    //=============================================================================
    //=============================================================================
    void SharedData::stop_thread()
    {
        {
            omni_mutex_lock sync(*this);
            stop_it = true;
        }
        signal();
        //condition.signal();
    }
    //=============================================================================
    //=============================================================================
    void SharedData::wait_initialized()
    {
        init_mutex.lock();
        while(!initialized)
        {
            init_condition.wait();
        }
        init_mutex.unlock();
    }

    //=============================================================================
    //=============================================================================
    SubscribeThread::SubscribeThread(Tango::DeviceImpl *dev):Tango::LogAdapter(dev)
    {
        _device = dev;
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        period  = _dev->subscribePeriod;
        shared  = _dev->shared;
    }
    //=============================================================================
    //=============================================================================
    auto SubscribeThread::run_undetached(void * /*ptr*/) -> void *
    {
        AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
        INFO_STREAM << "SubscribeThread id="<<omni_thread::self()->id()<<endl;
        while(!shared->get_if_stop())
        {
            // Try to subscribe
            DEBUG_STREAM << __func__<<": AWAKE"<<endl;
			_dev->push_thread->awake_new_data();	//awake periodically new_data omni_condition in get_next_cmds
            _dev->alarm_receivers->get_alarm_list_db(_dev->handler_list_fix, _dev->alarm_list);
            for(auto it : _dev->alarm_list)
            {
                try
                {
                    shared->add(it);
                }
                catch(Tango::DevFailed &e)
                {
                    //INFO_STREAM << ": error adding " << it << ", err=" << e.errors[0].desc;
                }
            }
            //HERE loop alarms in error in shared, and remove alarms not present in _dev->alarm_list

            vector<string> checklist;
            shared->get_sig_list(checklist);
            for(auto it : checklist)
            {
                if (std::find(_dev->alarm_list.begin(), _dev->alarm_list.end(), it) == _dev->alarm_list.end())
                {
                    try
                    {
                        DEBUG_STREAM << __func__ << ": removing no more present " << it;
                        shared->remove(it, true);
                        shared->remove(it, false);
                    }
                    catch(Tango::DevFailed &e)
                    {
                        //INFO_STREAM << ": error removing " << it << ", err=" << e.errors[0].desc;
                    }
                }
            }
            shared->start_all();
            shared->subscribe_events();
            int nb_to_subscribe = shared->nb_sig_to_subscribe();
            // And wait a bit before next time or
            // wait a long time if all signals subscribed
            {
                omni_mutex_lock sync(*shared);
                {
                    DEBUG_STREAM << __func__<<": going to wait period="<<period<<"  nb_to_subscribe="<<nb_to_subscribe<<endl;
                    //unsigned long s,n;
                    //omni_thread::get_time(&s,&n,period,0);
                    //shared->condition.timedwait(s,n);
                    shared->wait(std::chrono::milliseconds(std::chrono::seconds(period)).count());
                }
                //shared->unlock();
            }
        }
        shared->clear_signals();
        INFO_STREAM <<"SubscribeThread::"<< __func__<<": exiting..."<<endl;
        return nullptr;
    }
    //=============================================================================
    //=============================================================================



} // namespace
